from typing import List

from pyllk.production_rule import ProductionRule


class Grammar:
    '''There is only support for LL(1) grammars in this prototype version'''

    production_rules: List
    lookup_table = {}

    def __init__(self, production_rules: List[ProductionRule]):
        for pr in production_rules:
            if pr.source.name.lower() == 'init':
                break
        else:
            raise Exception('NO_INIT', 'Missing INIT production rule')

        self.production_rules = production_rules
        self.lookup_table = {}

        for pr in production_rules:
            items = self.lookup_table[pr.source.name] if pr.source.name in self.lookup_table else []
            items.append(pr)
            self.lookup_table[pr.source.name] = items



from typing import Dict, List

from pyllk.token import NonTerminalToken, Token


class ProductionRule:
    source: NonTerminalToken
    produced: List

    action = None
    ''' When specified the action is a function that takes am ExecutionContext as parameter'''

    def __init__(self, source: NonTerminalToken, produced: List, action=None, data={}):
        self.source = source
        self.produced = produced
        self.action = action
        self.data = data

    def __str__(self):
        produced = " ".join([p.__str__() for p in self.produced])
        return "{source}: {produced}".format(source=self.source, produced=produced)


class ExecutionContext:
    parser = None
    rule: ProductionRule
    tokens: List[Token]
    context: Dict
    consumed_tokens: List

    def __init__(self, parser, rule: ProductionRule, tokens: List[Token], context: Dict, consumed_tokens: []):
        self.parser = parser
        self.rule = rule
        self.tokens = tokens
        self.context = context
        self.consumed_tokens = consumed_tokens



import copy
import json
from typing import Dict

from pyllk.grammar import Grammar
from pyllk.production_rule import ExecutionContext, ProductionRule
from pyllk.token import NonTerminalToken, TerminalToken
from pyllk.token_stream import TokenStream


class Parser:
    grammar: Grammar
    debug = False

    def __init__(self, grammar: Grammar, debug=False):
        self.grammar = grammar
        self.debug = debug

    def parse(self, token_stream: TokenStream, context: Dict = {}):
        matched = self.consume_next_sub_rule(NonTerminalToken("INIT"), token_stream, context)
        return False if matched is None else True

    def parse_string(self, input: str, context: Dict = {}):
        '''Convenience method for parsing strings'''
        self.log(input)
        return self.parse(TokenStream([TerminalToken(c) for c in input]), context)

    def log(self, message: str):
        if self.debug:
            print(message)

    def consume_next_sub_rule(self, from_state: NonTerminalToken, token_stream: TokenStream, context: Dict = {}):
        if from_state.name not in self.grammar.lookup_table:
            raise Exception('UNKNOWN_STATE', 'The given state name {} is invalid'.format(from_state.name))

        rules = self.grammar.lookup_table[from_state.name]

        for rule in rules:
            self.log("Check rule({}) {}".format(token_stream.head, rule))
            consumed_tokens = self.consume(rule, token_stream, context)
            if consumed_tokens is not None:
                self.log("Consumed rule({}) {}".format(token_stream.head, rule))
                return consumed_tokens

    def consume(self, rule: ProductionRule, token_stream: TokenStream, context: Dict = {}):
        '''Tries to recursively consume a production rule. Returns False
        if it cannot do it and leaves the token stream at the position it
        initally found it by back-tracking.
        '''

        reset_head = token_stream.head
        reset_context = copy.deepcopy(context)
        self.log(json.dumps(reset_context))

        consumed_tokens = []

        for token in rule.produced:
            if isinstance(token, TerminalToken):
                if token.is_empty():
                    pass  # Empty matches every token
                elif (not token_stream.is_drained()) and (token.matches(token_stream.peek())):
                    tk = token_stream.consume()
                    consumed_tokens.append(tk)
                    self.log("Consumed token({}): {}".format(token_stream.head - 1, tk))
                else:
                    if self.debug:
                        if token_stream.is_drained():
                            self.log("Reset token stream to {} because drained stream".format(reset_head))
                        else:
                            self.log("Reset token stream to {} because of unmatched token: '{}'. Expected '{}'".format(reset_head, token_stream.peek(), token))

                    token_stream.set_to(reset_head)
                    context.clear()
                    for key in reset_context:
                        context[key] = reset_context[key]

                    return None
            else:
                sub_consumed_tokens = self.consume_next_sub_rule(token, token_stream, context)
                if not sub_consumed_tokens:
                    self.log("Reset token stream to {}".format(reset_head))
                    token_stream.set_to(reset_head)
                    context.clear()
                    for key in reset_context:
                        context[key] = reset_context[key]

                    return None
                
                consumed_tokens.append(sub_consumed_tokens)
                

        if rule.action:
            # Rule is about to be accepted, execute the action if any.
            # Give the action the list of tokens it consumed
            self.log("EXECUTE({}) for {}".format(token_stream.head, rule))
            rule.action(ExecutionContext(self, rule, token_stream.tokens[reset_head:token_stream.head], context, consumed_tokens))

        return consumed_tokens


class Context(dict):
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)



import json

from pyllk.token import TerminalToken


class SpacyToken(TerminalToken):
    def __init__(self):
        super().__init__(representation={})

    def text(self, text):
        self.representation['TEXT'] = text
        return self

    def lemma(self, text):
        self.representation['LEMMA'] = text
        return self

    def dependency(self, text):
        self.representation['DEP'] = text
        return self

    def part_of_speech(self, text):
        self.representation['POS'] = text
        return self

    def tag(self, text):
        self.representation['TAG'] = text
        return self

    def entity(self, text):
        self.representation['ENTITY'] = text
        return self

    def regex(self, text):
        self.representation['REGEX'] = text
        return self

    def matches(self, obj):
        match = True

        if 'TEXT' in self.representation:
            match &= (self.representation['TEXT'] == obj.text)

        if 'LEMMA' in self.representation:
            match &= (self.representation['LEMMA'] == obj.lemma_)

        if 'DEP' in self.representation:
            match &= (self.representation['DEP'] == obj.dep_)

        if 'POS' in self.representation:
            match &= (self.representation['POS'] == obj.pos_)

        if 'TAG' in self.representation:
            match &= (self.representation['TAG'] == obj.tag_)

        if 'ENTITY' in self.representation:
            match &= (self.representation['ENTITY'] == obj.ent_type_)

        return match

    def __str__(self):
        return json.dumps(self.representation)



import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='pyllk',
    version='0.0.8',
    description='A python LL(k) parser with a twist where tokens can be any arbitrary objects.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    keywords='parser ll(k) lex yacc',
    url='https://gitlab.com/majorspot/libraries/pyllk',
    author='Michael Privat',
    author_email='mprivat@majorspot.com',
    license='MIT',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(exclude=['*.tests']))



from typing import List


class TokenStream:
    head: int
    tokens: List

    def __init__(self, tokens: List):
        self.head = 0
        self.tokens = tokens

    def consume(self):
        result = self.tokens[self.head]
        self.head += 1
        return result

    def left_overs(self):
        return self.tokens[self.head:]

    def is_drained(self):
        return self.head == len(self.tokens)

    def set_to(self, head: int):
        self.head = head

    def peek(self):
        return self.tokens[self.head]

    @classmethod
    def from_iterator(cls, it):
        tokens_list = []
        for token in it:
            tokens_list.append(token)

        return TokenStream(tokens_list)



class Token:
    pass


class TerminalToken(Token):
    representation: object

    def __init__(self, representation: object):
        self.representation = representation

    def __eq__(self, obj):
        return self.representation == obj

    def __str__(self):
        return self.representation

    def matches(self, obj):
        return self == obj

    def is_empty(self):
        '''Override this method to provide a different implementation of empty
        '''
        return self == ''


class NonTerminalToken(Token):
    name: str

    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return "[{}]".format(self.name)
